from decimal import Decimal
from typing import Dict, List

import pdfplumber

from disclosure_extractor.jef.filters import add_lines


def get_lines(page):
    filtered_page = page.filter(add_lines).extract_words()
    lines = page.lines
    keys = []
    for f in filtered_page:
        f["object_type"] = "line"
        f["stroking_color"] = (1, 2, 3)
        f["non_stroking_color"] = None
        f["x1"] = page.width
        f["x0"] = 0
        f["linewidth"] = Decimal("2")
        f["width"] = 600
        f["doctop"] = f["top"] + 700
        f["top"] = f["top"] + 20
        keys.append(f["top"])
        lines.append(f)
    return lines, keys


def set_section(t):
    dc = {
        "I. Positions": [
            "#",
            "ORGANIZATION NAME",
            "NOTE",
            "ORGANIZATION TYPE",
            "POSITION HELD",
        ],
        "II. Agreements": ["#", "EMPLOYER OR PARTY", "NOTE", "TERMS", "DATE"],
        "III A. Filer's Non-Investment Income": [
            "#",
            "SOURCE",
            "NOTE",
            "INCOME TYPE",
            "INCOME AMOUNT",
        ],
        "III B. Spouse's Non-Investment Income": [
            "#",
            "SOURCE",
            "NOTE",
            "DESCRIPTION",
        ],
        "IV. Reimbursements": [
            "#",
            "SOURCE",
            "NOTE",
            "DATES",
            "LOCATION",
            "PURPOSE",
            "ITEMS PAID OR PROVIDED",
        ],
        "V. Gifts": ["#", "SOURCE", "NOTE", "DESCRIPTION", "VALUE"],
        "VI. Liabilities": ["#", "CREDITOR", "NOTE", "TYPE", "VALUE"],
        "VII. Investments and Trusts": [
            "DESCRIPTION",
            "NOTE",
            "INCOME AMOUNT",
            "INCOME TYPE",
            "GROSS VALUE AS OF 12/31",
            "VALUE METHOD",
            "TRANS. TYPE",
            "TRANS. DATE",
            "TRANS. VALUE",
        ],
        "Additional Information or Explanation": ["PART", "#", "NOTE"],
    }
    section_titles = {
        "III A. Filer's Non-Investment Income": "Non-Investment Income",
        "III B. Spouse's Non-Investment Income": "Non Investment Income Spouse",
        "Additional Information or Explanation": ["PART", "#", "NOTE"],
    }
    if "III" in t:
        title = section_titles[t]
    elif "Additional" in t:
        title = "Additional Information or Explanations"
    else:
        title = t.split(". ")[1]

    return dc[t], title


def crop_and_extract(
    page: pdfplumber.pdf.Page,
    line: Dict,
    adjust=False,
    left_shift: int = 0,
    up_shift: int = 12,
    keys: List = [],
) -> str:
    """Extract text content for pdf line if any

    Given a line of a pdf - extracted the text around it.  If adjust
    is True, reduce the cropped area to within the first line (usually above)
    :param page: Page to crop
    :param line: Line to crop around
    :param adjust: Whether to check if another line is inside our crop
    :param left_shift: Leftward crop adjustment
    :param up_shift: Upward crop adjustment
    :return: Content of the section
    """
    bbox = (
        int(line["x0"]) - left_shift,
        max(int(line["top"]) - up_shift, 0),
        int(line["x1"]),
        int(line["top"]),
    )
    crop = page.crop(bbox)
    if adjust:
        tops = [row["top"] for row in crop.lines if row["top"] != line["top"]]
        if len(tops) > 0:
            nbb = (*bbox[:1], tops[-1], *bbox[2:])
            for i in keys:
                if nbb[1] < i < nbb[3]:
                    bbox = (*bbox[:1], int(i), *bbox[2:])

                    # return nbb
            # crop = page.filter(regular_text).crop(bbox=nbb)

            bbox = (*bbox[:1], int(tops[-1]), *bbox[2:])

    return bbox



from decimal import Decimal


def title_text(obj):
    """
    Filter for bold text in Times-Bold font
    """
    if (
        obj.get("object_type", None) == "char"
        and obj.get("size", None) in [10, 12]
        and obj.get("fontname", None) == "Times-Bold"
    ):
        return True
    return False


def title_text_reverse(obj):
    """
    Filter for bold text in Times-Bold font
    """
    if (
        obj.get("object_type", None) == "char"
        and obj.get("size", None) == 12
        and obj.get("fontname", None) == "Times-Bold"
    ):
        return False
    return True


def filter_bold_times(obj):
    """
    Filter for bold text in Times-Bold font
    """
    if obj.get("fontname", None) == "Times-Bold" or obj.get("size", 0) < 8:
        return False
    return True


def title_text_V(obj):
    """
    Filter for bold text in Times-Bold font
    """
    if (
        obj.get("object_type", None) == "char"
        and obj.get("size", None) == 12
        and obj.get("fontname", None) == "Times-Bold"
        and obj.get("text", None) == "V"
    ):
        return True
    return False


def filter_bold(obj):
    if (
        obj["object_type"] == "char"
        and obj["fontname"] == "YUQOTN+OpenSans-Semibold"
    ):
        return True
    elif obj["object_type"] == "char" and obj["fontname"] == "UZGVXX+OpenSans":
        return True
    return False


def header_text(obj):
    if obj["object_type"] == "char" and obj["size"] == 14:
        return True


def bold_lines(obj):
    if obj["object_type"] == "line" and obj["linewidth"] == Decimal("2"):
        return True


def regular_text(obj):
    if obj["object_type"] == "char" and obj["size"] == 14:
        return False
    return True


def add_lines(obj):
    if (
        obj["object_type"] == "char"
        and int(obj["size"]) == 14
        and obj["text"] == "."
    ):
        return True
    if (
        obj["object_type"] == "char"
        and int(obj["size"]) == 14
        and obj["text"] == "x"
    ):
        return True
    return False



# -*- coding: utf-8 -*-

import json
import logging
from itertools import groupby
from typing import Dict, Union, List

import cv2
import numpy as np
import pandas as pd
from PIL.Image import Image

try:
    import importlib.resources as importlib_resources
except ImportError:
    # In PY<3.7 fall-back to backported `importlib_resources`.
    import importlib_resources


def clahe(img, clip_limit=1.0, grid_size=(8, 8)):
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=grid_size)
    return clahe.apply(img)


def determine_section_of_contour(checkboxes, rect):
    """This does work with my hacky solution.
    this needs work - but afterwards we should be in good shape to finish this project
    wanna be bigger than checkbox so sort by y - normally- and as soon as we are larger than
    """
    search = [x for x in checkboxes if x[4] == rect[4]]
    for y in sorted(search, key=lambda x: (x[1]))[::-1]:
        if rect[1] > y[1]:
            return y[5]
    return max([x[5] for x in checkboxes])


def load_template():
    f = importlib_resources.read_text(
        "disclosure_extractor", "extractor_template.json"
    )
    return json.loads(f)


def erode(image, q):
    """

    :param image:
    :param q:
    :return:
    """
    kernel = np.ones((q, q), np.uint8)
    return cv2.erode(image, kernel, iterations=1)


def process_contours_page(
    page_image: Image,
    results: Dict[str, Union[str, int, float, List, Dict]],
    pg_num: int,
    checkboxes: List,
    check: Dict,
    s0: List,
    s7: List,
    little_checkboxes: List,
    s1: List,
    try_again: bool,
) -> Dict[str, Union[str, int, float, List, Dict]]:

    # Add to queue

    mode = cv2.RETR_CCOMP
    method = cv2.CHAIN_APPROX_SIMPLE
    cv_image = np.array(page_image)
    image = process_image(page_image)
    height, width = image.shape
    contours, hierarchy = cv2.findContours(image, mode, method)

    # Obtain the checkboxes on the page- to determine what section we are processing
    # do this first so we can remove any noise we might bump into on the top of the page

    i = len(contours) - 1
    checkboxes_on_page = []
    while 0 < i:
        cntr = contours[i]
        area = cv2.contourArea(cntr)
        x, y, w, h = cv2.boundingRect(cntr)
        rect_area = w * h
        extent = float(area) / rect_area
        aspect_ratio = float(w) / h
        if 0.9 <= aspect_ratio <= 1.1 and extent > 0.8 and x < width * 0.2:
            if not try_again:
                move_on = True if hierarchy[0, i, 3] == -1 else False
            else:
                move_on = True if hierarchy[0, i, 3] > -1 else False
            if move_on:
                checkboxes_on_page.append((x, y, w, h, pg_num))
                sect_order = (
                    len(checkboxes) + 1 if len(checkboxes) + 1 < 9 else 8
                )
                mean = (
                    sum(
                        np.array(
                            cv2.mean(cv_image[y : y + h, x : x + w])
                        ).astype(np.uint8)
                    )
                    / 3
                )
                is_empty = False
                if mean < 230:
                    is_empty = True

                for k, sect in results["sections"].items():
                    if sect["order"] == sect_order:
                        section = k

                checkboxes.append(
                    (
                        x,
                        y,
                        w,
                        h,
                        pg_num,
                        section,
                        {"is_section_empty": is_empty, "mean": mean},
                    )
                )
                check[section] = is_empty
        i -= 1
    i = 0
    min_y = 0
    if checkboxes_on_page:
        min_y = min([x[1] for x in checkboxes_on_page])

    while i < len(contours):
        cntr = contours[i]
        area = cv2.contourArea(cntr)
        x, y, w, h = cv2.boundingRect(cntr)
        rect_area = w * h
        extent = float(area) / rect_area
        aspect_ratio = float(w) / h

        # Date and name information
        if (
            y < height * 0.5 and h > 50 and x > width * 0.5 and pg_num == 0
        ) or (height * 0.05 < y < height * 0.1 and h > 50 and pg_num == 0):
            if hierarchy[0, i, 3] == -1:
                # cv2.drawContours(cv_image, contours, i, (255, 70, 70))
                s0.append((x, y, w, h))

        # Cells for Investments and Trusts  √√√√√
        # upper = 180 if height > 3000 else 80
        # print(w, h, len(checkboxes))
        if 10 > w / h > 0.9 and 150 > h > 40 and len(checkboxes) > 7:
            # This lets me remove overlapping boxes,
            # and take the inner, cleaner version
            if hierarchy[0, i, 3] == -1:
                # cv2.drawContours(cv_image, contours, i, (70, 70, 255))
                s7.append(
                    (
                        x,
                        y,
                        w,
                        h,
                        pg_num,
                        range(y, y + h),
                        "Investments and Trusts",
                    )
                )

        # Highlight text input lines  √√√√√√
        if w / h > 7 and w > 150 and y > min_y:
            # if hierarchy[0, i, 3] == -1:
            rect = (x, y, w, h, pg_num, range(y, y + h))
            if not checkboxes:
                return results
            section = determine_section_of_contour(checkboxes, rect)
            # section = section_dict[section]['name']
            rect = (x, y, w, h, pg_num, range(y, y + h), section)
            s1.append(rect)
            # cv2.drawContours(cv_image, contours, i, (70, 70, 255))
        # Find small checkboxes on first page  ** this need to be completed
        if (
            0.9 <= aspect_ratio <= 1.1
            and extent > 0.8
            and x > width * 0.2
            and 50 > h > 20
            and pg_num == 0
        ):
            if hierarchy[0, i, 3] == -1:
                # Process the little red checkbox at the start
                # cv2.drawContours(cv_image, contours, i, (0, 0, 255))
                mean = (
                    sum(
                        np.array(
                            cv2.mean(cv_image[y : y + h, x : x + w])
                        ).astype(np.uint8)
                    )
                    / 3
                )
                little_checkboxes.append((x, y, w, h, mean))
                # font = cv2.FONT_HERSHEY_SIMPLEX
                # cv2.putText(cv_image, str(int(mean)), (x, y), font, 1,
                #             (0, 255, 0), 2, cv2.LINE_AA)

        i += 1

    if pg_num == 0 and len(little_checkboxes) == 5:
        sorted_little_checkboxes = sorted(
            little_checkboxes, key=lambda x: x[1]
        )
        nomination = sorted_little_checkboxes.pop(0)
        amended = sorted_little_checkboxes.pop(-1)
        initial, annual, final = sorted(
            sorted_little_checkboxes, key=lambda x: (x[0])
        )
        results["nomination"] = True if nomination[4] < 220 else False
        results["amended"] = True if amended[4] < 220 else False
        results["initial"] = True if initial[4] < 220 else False
        results["annual"] = True if annual[4] < 220 else False
        results["final"] = True if final[4] < 220 else False

    # Release from queue
    logging.info("Page contours extracted")
    return results


def process_i_row(results, investments, check):
    """

    :param results:
    :param investments:
    :param check:
    :return:
    """
    investment_group = groupby(
        investments["data"],
        lambda content: content["group"],
    )

    row_index = 0
    for grouping in investment_group:
        col_indx = 0
        groups = list(grouping[1])
        results["sections"]["Investments and Trusts"]["rows"][row_index] = {}
        for group in sorted(groups, key=lambda x: x["x"]):
            group["coords"] = (
                group["x"],
                group["y"],
                (group["x"] + group["w"]),
                (group["y"] + group["h"]),
            )
            column = results["sections"]["Investments and Trusts"]["fields"][
                col_indx
            ]
            results["sections"]["Investments and Trusts"]["rows"][row_index][
                column
            ] = group
            results["sections"]["Investments and Trusts"]["empty"] = check[
                "Investments and Trusts"
            ]
            col_indx += 1
        row_index += 1

    # Remove any incomplete rows
    irows = results["sections"]["Investments and Trusts"]["rows"]
    for r_index in irows:
        if len(irows[r_index].keys()) != 10:
            del irows[r_index]
    return results


def extract_other_data(results, check, other_sections):
    """

    @param results:
    @param check:
    @param other_sections:
    @return:
    """
    other_groups = groupby(
        other_sections["data"],
        lambda content: content["group"],
    )

    row_index = 0
    for grouping in other_groups:
        col_indx = 0
        groups = list(grouping[1])
        sect = groups[0]["section"]
        if results["sections"][sect]["rows"] == {}:
            row_index = 0
        results["sections"][sect]["rows"][row_index] = {}
        if len(groups) != len(results["sections"][sect]["fields"]):
            continue
        if sorted(groups, key=lambda x: x["x"])[0]["x"] > 120:
            continue
        for group in sorted(groups, key=lambda x: x["x"]):
            group["coords"] = (
                group["x"],
                group["y"] - 60,
                (group["x"] + group["w"]),
                (group["y"] + group["h"]),
            )
            try:
                column = results["sections"][sect]["fields"][col_indx]
                results["sections"][sect]["rows"][row_index][column] = group
                results["sections"][sect]["empty"] = check[sect]
                col_indx += 1
            except:
                pass
        row_index += 1
    return results


def group_together_investments(s7):
    """

    :param s7:
    :return:
    """
    df = pd.DataFrame(
        {
            "x": [x[0] for x in s7],
            "y": [x[1] for x in s7],
            "w": [x[2] for x in s7],
            "h": [x[3] for x in s7],
            "top": [x[1] + 10 for x in s7],
            "page": [x[4] for x in s7],
            "section": "Investments and Trusts",
        }
    )
    df = df.sort_values(["page", "y", "top"])
    df["group"] = (
        (
            df.top.rolling(window=2, min_periods=1).min()
            - df.y.rolling(window=2, min_periods=1).max()
        )
        < 0
    ).cumsum()
    ndf = df.groupby("group").filter(lambda x: len(x) == 10)
    investments = json.loads(ndf.to_json(orient="table"))
    return investments


def group_other_sections(s1):
    """

    @param s1:
    @return:
    """
    df2 = pd.DataFrame(
        {
            "x": [x[0] for x in s1],
            "y": [x[1] for x in s1],
            "w": [x[2] for x in s1],
            "h": [x[3] for x in s1],
            "top": [x[1] + 10 for x in s1],
            "page": [x[4] for x in s1],
            "section": [x[6] for x in s1],
        }
    )
    df2 = df2.sort_values(["page", "y", "top"])
    df2["group"] = (
        (
            df2.top.rolling(window=2, min_periods=1).min()
            - df2.y.rolling(window=2, min_periods=1).max()
        )
        < 0
    ).cumsum()

    df2 = df2[df2.section != "Investments and Trusts"]
    ndf2 = df2.groupby("group").filter(lambda x: len(x) > 1)
    other_sections = json.loads(ndf2.to_json(orient="table"))
    return other_sections


def extract_contours_from_page(pages: List[Image], try_again):
    """Process PDF

    Return the document structure as JSON data to easily and accurately
    extract out the information.
    """
    results = load_template()
    check = {}
    checkboxes, s0, s1, s7 = [], [], [], []
    little_checkboxes = []
    pg_num = 0
    for page in pages:
        results = process_contours_page(
            page,
            results,
            pg_num,
            checkboxes,
            check,
            s0,
            s7,
            little_checkboxes,
            s1,
            try_again,
        )
        pg_num += 1

    # Here is where we do some data processing and group rows together
    # sometimes things need to be massaged a bit.

    # Extract investment table data
    investments = group_together_investments(s7)
    results = process_i_row(results, investments, check)

    # Extract sections I to VI
    other_sections = group_other_sections(s1)
    results = extract_other_data(results, check, other_sections)

    results["first_four"] = s0
    results["page_count"] = len(pages)
    results["found_count"] = len(checkboxes)
    return results


def process_image(input_image: Image) -> Image:
    """

    @param input_image:
    @return:
    """
    cv_image = np.array(input_image)
    src = cv_image[:, :, ::-1].copy()

    # HSV thresholding to get rid of as much background as possible
    hsv = cv2.cvtColor(src.copy(), cv2.COLOR_BGR2HSV)
    mask = cv2.inRange(
        src=hsv,
        lowerb=np.array([0, 0, 0]),
        upperb=np.array([255, 255, 255]),
    )
    result = cv2.bitwise_and(src, src, mask=mask)
    _, gg, _ = cv2.split(result)
    gg = clahe(gg, 1, (3, 3))

    # Adaptive Thresholding to isolate the bed
    image = cv2.adaptiveThreshold(
        src=cv2.blur(gg, (1, 1)),
        maxValue=255,
        adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        thresholdType=cv2.THRESH_BINARY,
        blockSize=41,  # maybe variable...
        C=2,
    )
    return image


def find_redactions(image_crop: Image) -> bool:
    """Find redactions

    If boxes are found inside an image crop - assume redactions occurred.

    :param image_crop: Image crop
    :return: Whether the image contains redactions
    """
    mode = cv2.RETR_CCOMP
    method = cv2.CHAIN_APPROX_SIMPLE
    image = process_image(image_crop)
    contours, hierarchy = cv2.findContours(image, mode, method)

    i = 0
    while i < len(contours):
        cntr = contours[i]
        area = cv2.contourArea(cntr)
        x, y, w, h = cv2.boundingRect(cntr)
        rect_area = w * h
        extent = float(area) / rect_area
        aspect_ratio = float(w) / h
        if 0.9 <= aspect_ratio <= 10.1 and extent > 0.8 and 50 > h > 20:
            return True
        i += 1
    return False


class Error(Exception):
    """Base class for other exceptions"""

    pass


class CheckboxesNotFound(Error):
    """Raised when contour for checkboxes could not be determined"""

    pass


def clean_image(image_crop: Image) -> Image:
    """Remove table lines that could be part of the image crop

    @param image_crop:
    @return: Cleaned Image
    """
    image = np.array(image_crop)
    # Convert RGB to BGR
    image = image[:, :, ::-1].copy()

    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    thresh = cv2.threshold(
        gray, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
    )[1]

    # Remove horizontal
    horizontal_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (10, 1))
    detected_lines = cv2.morphologyEx(
        thresh, cv2.MORPH_OPEN, horizontal_kernel, iterations=2
    )
    cnts = cv2.findContours(
        detected_lines, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    cnts = cnts[0] if len(cnts) == 2 else cnts[1]
    image_h, image_w, _ = image.shape
    i = 0
    for c in cnts:
        x, y, w, h = cv2.boundingRect(c)
        if image_h - 10 > y > 10:
            i += 1
            continue
        cv2.drawContours(image, cnts, i, (255, 255, 255), 2)
        i += 1

    # Remove vertical
    vertical_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 10))
    detected_lines = cv2.morphologyEx(
        thresh, cv2.MORPH_OPEN, vertical_kernel, iterations=2
    )
    cnts = cv2.findContours(
        detected_lines, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    cnts = cnts[0] if len(cnts) == 2 else cnts[1]
    for c in cnts:
        cv2.drawContours(image, [c], -1, (255, 255, 255), 2)

    # Remove whitespace to improve tesseract
    # image = cv2.imread(f)
    original = image.copy()
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    blur = cv2.GaussianBlur(gray, (25, 25), 0)
    thresh = cv2.threshold(
        blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
    )[1]

    # Perform morph operations, first open to remove noise, then close to combine
    noise_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
    opening = cv2.morphologyEx(
        thresh, cv2.MORPH_OPEN, noise_kernel, iterations=2
    )
    close_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (7, 7))
    close = cv2.morphologyEx(
        opening, cv2.MORPH_CLOSE, close_kernel, iterations=3
    )

    # Find enclosing bounding box and crop ROI
    coords = cv2.findNonZero(close)
    x, y, w, h = cv2.boundingRect(coords)
    cv2.rectangle(image, (x, y), (x + w, y + h), (36, 255, 12), 2)
    crop = original[y : y + h, x : x + w]

    return crop



import os
import re
from typing import Dict, Tuple

import pdfplumber

from disclosure_extractor.image_processing import load_template
from disclosure_extractor.jef.filters import (
    filter_bold_times,
    title_text,
    title_text_V,
    title_text_reverse,
)
from disclosure_extractor.jef.utils import (
    crop_and_extract,
    get_lines,
    set_section,
)

VALUATION_METHODS = (
    ("Q", "Appraisal"),
    ("R", "Cost (Real Estate Only)"),
    ("S", "Assessment"),
    ("T", "Cash Market"),
    ("U", "Book Value"),
    ("V", "Other"),
    ("W", "Estimated"),
)

INCOME_GAIN = (
    ("A", "1 - 1,000"),
    ("B", "1,001 - 2,500"),
    ("C", "2,501 - 5,000"),
    ("D", "5,001 - 15,000"),
    ("E", "15,001 - 50,000"),
    ("F", "50,001 - 100,000"),
    ("G", "100,001 - 1,000,000"),
    ("H1", "1,000,001 - 5,000,000"),
    ("H2", "5,000,001 +"),
)

GROSS_VALUE = (
    ("J", "1 - 15,000"),
    ("K", "15,001 - 50,000"),
    ("L", "50,001 - 100,000"),
    ("M", "100,001 - 250,000"),
    ("N", "250,001 - 500,000"),
    ("O", "500,001 - 1,000,000"),
    ("P1", "1,000,001 - 5,000,000"),
    ("P2", "5,000,001 - 25,000,000"),
    ("P3", "25,000,001 - 50,000,000"),
    ("P4", "50,000,001 - "),
)


def get_comment(content) -> str:
    """Check if comments are found on the first page - this is a new comment

    This new general comment is appended to the additional comments section.

    :param content:
    :return:
    """
    if "COMMENT" in content:
        comment = content.split("COMMENT")[1]
        return comment.replace("\n", " ").strip()
    return ""


def get_metadata(cd, content) -> Tuple:
    """Check if comments are found on the first page, and return them

    :param page:
    :return:
    """

    cd["report_type"] = content.split(" ")[0]
    cd["year"] = re.findall(r"\d{4}", content)[0]
    cd["judge"] = re.findall(r"\/s\/(.*)\[", content)[0].strip()
    cd["annual"] = True if "Annual" in cd["report_type"] else False
    cd["date_of_report"] = re.findall(r"signed on(.*)by", content)[0].strip()
    return cd


def convert_gross_value(datum):
    """Convert gross values to code for gross values

    :param datum:
    :return:
    """
    if datum == None or datum == "None":
        return ""

    if datum:
        if "less" in datum:
            return "J"
        min_val = datum.split(" ")[0]
        h = re.sub(r"[^\d,]", "", min_val)
        for k, v in GROSS_VALUE:
            if v.split(" ")[0] == h:
                return k
    return datum


def convert_valuation_method(datum):
    """Convert values to code for Valuation method

    :param datum:
    :return:
    """
    if datum == None or datum == "None":
        return ""

    if datum:
        for k, v in VALUATION_METHODS:
            if v == datum:
                return k
    return datum


def convert_income_gain(datum):
    """Convert values to code for income gains

    :param datum:
    :return:
    """
    if datum == None or datum == "None":
        return ""

    if datum:
        if datum == "$1,000 or less":
            return "A"
        min_val = datum.split(" ")[0]
        h = re.sub(r"[^\d,]", "", min_val)
        for k, v in INCOME_GAIN:
            if v.split(" ")[0] == h:
                return k
    return datum


def add_data(results: Dict, title: str, data: Dict, pg_number: int) -> Dict:
    """Normalize and combine data into results

    :param results: All extracted content
    :param title: Section title
    :param data: Row data to append to results
    :param pg_number: page number
    :return: Results with appended data row
    """
    for k, v in data.items():
        if v == "None" or v == None:
            data[k] = ""

    if not results["sections"][title]["rows"]:
        results["sections"][title]["rows"] = []

    if title == "Positions":
        r = {
            "Name of Organization": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["ORGANIZATION NAME"]} - {data["ORGANIZATION TYPE"]}',
            },
            "Position": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["POSITION HELD"],
            },
        }
    elif title == "Agreements":
        r = {
            "Date": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["DATE"]}',
            },
            "Parties and Terms": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["EMPLOYER OR PARTY"]} - {data["TERMS"]}',
            },
        }
    elif title == "Non-Investment Income":
        r = {
            "Date": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": "",
            },
            "Source and Type": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["SOURCE"]} - {data["INCOME TYPE"]}',
            },
            "Income": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["INCOME AMOUNT"]}',
            },
        }
    elif title == "Non Investment Income Spouse":
        r = {
            "Date": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": "",
            },
            "Source and Type": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["SOURCE"]} - {data["DESCRIPTION"]}',
            },
        }
    elif title == "Reimbursements":
        r = {
            "Source": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["SOURCE"]}',
            },
            "Dates": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["DATES"]}',
            },
            "Locations": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["LOCATION"]}',
            },
            "Purpose": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["PURPOSE"]}',
            },
            "Items Paid or Provided": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": f'{data["ITEMS PAID OR PROVIDED"]}',
            },
        }
    elif title == "Gifts":
        r = {
            "Source": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["SOURCE"],
            },
            "Description": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["DESCRIPTION"],
            },
            "Value": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["VALUE"],
            },
        }
    elif title == "Liabilities":
        r = {
            "Creditor": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["CREDITOR"],
            },
            "Description": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["TYPE"],
            },
            "Value Code": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": convert_gross_value(data["VALUE"]),
            },
        }
    elif title == "Investments and Trusts":
        r = {
            "A": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["DESCRIPTION"],
                "inferred_value": False,
            },
            "B1": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": convert_income_gain(data["INCOME AMOUNT"]),
            },
            "B2": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["INCOME TYPE"],
            },
            "C1": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": convert_gross_value(data["GROSS VALUE AS OF 12/31"]),
            },
            "C2": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": convert_valuation_method(data["VALUE METHOD"]),
            },
            "D1": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["TRANS. TYPE"],
            },
            "D2": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": data["TRANS. DATE"],
            },
            "D3": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": convert_gross_value(data["TRANS. VALUE"]),
            },
            "D4": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": "",
            },
            "D5": {
                "is_redacted": data["is_redacted"],
                "page_number": pg_number,
                "text": "",
            },
        }
    else:
        return results

    results["sections"][title]["rows"].append(r)
    return results


def extract_content(filepath: str) -> Dict:
    """Extract content from JEF financial disclosures.

    This functions works on a new type of document from the AO

    :param filepath: Path to the PDF
    :return: Extracted content
    :type: dict
    """
    with pdfplumber.open(filepath) as pdf:
        print(filepath)
        page_one = pdf.pages[0].extract_text()
        cd = load_template()

        cd = get_metadata(cd, page_one)

        yet = False
        addendum_redacted = False
        cd["general_comment"] = get_comment(page_one)
        cd["pdf_size"] = os.path.getsize(filepath)
        cd["page_count"] = len(pdf.pages)

        for page in pdf.pages:
            lines, keys = get_lines(page)
            row = []
            redacted = False

            for line in sorted(lines, key=lambda x: x["top"]):
                if line["width"] == 600:
                    bbox = crop_and_extract(
                        page, line, up_shift=25, adjust=False, keys=keys
                    )
                    output = page.crop(bbox=bbox).extract_text()
                    options, title = set_section(output)
                    row = []
                else:
                    bbox = crop_and_extract(
                        page, line, up_shift=150, adjust=True, keys=keys
                    )
                    for key in keys:
                        if int(bbox[1]) < int(key) < int(bbox[3]):
                            bbox = (*bbox[:1], int(key), *bbox[2:])
                    output = page.crop(bbox=bbox).extract_text()
                    if not redacted:
                        redacted = (
                            True if page.crop(bbox=bbox).curves else False
                        )
                    if output:
                        output = output.replace("\n", " ").strip()
                    if not yet and output == "#":
                        yet = True
                    if not yet:
                        continue
                    if output == None:
                        output = ""
                    row.append(output)
                    if row == ["None"]:
                        cd["sections"][title]["empty"] = True
                    elif len(row) == len(options):
                        if "#" in row or "DESCRIPTION" in row:
                            row = []
                            continue
                        data = dict(zip(options, row))
                        data["is_redacted"] = redacted
                        if title == "Additional Information or Explanations":
                            cd[
                                "general_comment"
                            ] = f'{cd["general_comment"]} PART {data["PART"]} #{data["#"]} {data["NOTE"]}'
                            if not redacted and data["is_redacted"]:
                                addendum_redacted = True
                            continue
                        cd["sections"][title]["empty"] = False
                        cd = add_data(cd, title, data, page.page_number)
                        row = []
                        redacted = False
        cd["Additional Information or Explanations"] = {
            "text": cd["general_comment"],
            "is_redacted": addendum_redacted,
        }
        return cd


def get_text(page, cell, field):
    """"""
    redacted = False
    if not cell:
        return
    crop = page.crop(cell).filter(filter_bold_times)
    if not crop:
        return
    text = crop.extract_text()
    # Check if yellow - and so not content
    if any([x["fill"] for x in crop.rects]):
        return
    if not text:
        text = ""

    if field == "":
        return text
    if crop.curves and crop.curves[0]["fill"]:
        redacted = True

    cleaned_text = re.sub(r"\s+", " ", text).strip("")

    if field in ["A", "Source", "Creditor", "Date", "Position"]:
        cleaned_text = re.sub(r"^\d{1,4}\.", "", cleaned_text).strip()

    if field == "A":
        cleaned_text = re.sub(r"^[-]{1,4}", "", cleaned_text).strip()

    return {
        "text": cleaned_text,
        "page_number": page.page_number,
        "is_redacted": redacted,
    }


def extract_normal_pdf(filepath: str) -> Dict:
    """"""

    titles = [
        "Positions",
        "Agreements",
        "Non-Investment Income",
        "Non Investment Income Spouse",
        "Reimbursements",
        "Gifts",
        "Liabilities",
        "Investments and Trusts",
    ]
    empty_template = load_template()

    with pdfplumber.open(filepath) as pdf:
        pages = pdf.pages
        for page in pages:
            if not page.extract_text() or len(page.extract_text()) < 100:
                return {
                    "success": False,
                    "msg": f"OCR required on page #{page.page_number}",
                }

        for page in pages:
            for table in page.debug_tablefinder().tables:
                c = get_text(page, table.cells[0], "")
                if c.startswith("1."):
                    title = titles.pop(0)
                    empty_template["sections"][title]["rows"] = []
                if not re.match(r"\d+\.", c):
                    continue
                content = empty_template["sections"][title]["rows"]
                for row in table.rows:
                    cd = {}
                    fields = empty_template["sections"][title]["fields"]
                    if len(row.cells) > len(fields):
                        cells = row.cells[1:]
                    else:
                        cells = row.cells
                    for cell, field in zip(cells, fields):
                        cd[field] = get_text(page, cell, field)

                    if any([v["text"] for k, v in cd.items()]):
                        content.append(cd)
                empty_template["sections"][title]["rows"] = content
        if len(titles) > 0:
            # Unable to find all tables for this modified
            empty_template["success"] = False
            return empty_template

        addendum = ""
        for page in pages:
            text = page.filter(title_text).extract_text()
            if not text:
                continue
            if "VIII. ADDITIONAL INFORMATION OR EXPLANATIONS." in text:
                top = page.filter(title_text_V).extract_words()[0]["bottom"]
                bbox = (0, top, page.width, page.height)
                crop = page.crop(bbox=bbox)
                text = crop.filter(title_text_reverse).extract_text()
                addendum = addendum + "\n " + text if text else ""

        empty_template["Additional Information or Explanations"][
            "text"
        ] = addendum

        empty_template["page_count"] = len(pdf.pages)
        empty_template["wealth"] = None
        empty_template["msg"] = ""
        empty_template["success"] = True

    count = 0
    investments = empty_template["sections"][title]["rows"]
    for investment_row in investments:
        investment_row["A"]["inferred_value"] = False
        if not count:
            count += 1
            continue
        investement_name = investment_row["A"]["text"]
        if (
            not investement_name
            and investments[count]["D1"]["text"]
            and investment_row["D1"]["text"]
        ):
            empty_template["sections"][title]["rows"][count]["A"][
                "inferred_value"
            ] = True
            empty_template["sections"][title]["rows"][count]["A"][
                "text"
            ] = investments[count - 1]["A"]["text"]
        count += 1

    return empty_template
